"""Parse domain names from provided list of repositories."""

from autogit.data_types import RepoState


def parse_domains(repos: dict[str, RepoState]) -> None:
    pass


# curl --header "PRIVATE-TOKEN: personal-access-token" your-gitlab-url/api/v4/version



from datetime import datetime
from unittest.mock import MagicMock

import pytest

from jsonstate import State


@pytest.mark.parametrize(
    ('state_name', 'initial_value', 'new_value'),
    [
        # String
        ('title', 'Initial title', 'Updated title'),
        ## list
        ('tags', ['Foo', 'Bar'], ['Eggs', 'Spam']),
        ## tuple
        ('tags', ('Foo', 'Bar'), ('Eggs', 'Spam')),
        ## dict
        ('metrics', {'speed': 1, 'excitment': 9}, {'saved_time': 99, 'joy': 99}),
        ## bool
        ('dark', False, True),
        ## int
        ('opacity', 99, 0),
        ## float
        ('value', 1.5, 79.8),
        ## datetime
        ('now', datetime(2025, 1, 30), datetime(2025, 2, 28)),
    ],
)
def test_state_observer_in_top_level(state_name, initial_value, new_value) -> None:
    """
    Test possible actions on top level keys.

    Interface of possible actions: (keyword only arguments)
    - action="add" new_value=<value>
    - action="update" old_value=<value>, new_value=<value>
    - action="remove" old_value=<value>
    """
    state_change_observer = MagicMock()

    state = State({state_name: initial_value})
    assert state
    assert state[state_name] == initial_value

    state.callbacks(key=state_name).append(
        state_change_observer
    )  # .on_change = state_change_observer
    assert state_change_observer in state.callbacks(key=state_name)

    state.pop(state_name)
    state_change_observer.assert_called_once_with(
        old_value=initial_value, action='remove'
    )
    state_change_observer.reset_mock()

    state[state_name] = initial_value
    state_change_observer.assert_called_once_with(new_value=initial_value, action='add')
    state_change_observer.reset_mock()

    state[state_name] = new_value
    state_change_observer.assert_called_once_with(
        new_value=new_value, old_value=initial_value, action='update'
    )
    state_change_observer.reset_mock()


def test_callback_for_dict_methods_at_the_top_level() -> None:
    """
    Test observer callback invocations for these Dict methods:
     - __setitem__
     - __delitem__
     - pop
     - popitem
     - clear
     - update - dict
     - update - iterable
     - update - kwargs
     - setdefault
    """
    state_change_observer = MagicMock()

    state = State({'foo': 'bar'})
    assert state
    assert state['foo'] == 'bar'

    state.callbacks(key='foo').append(
        state_change_observer
    )  # .on_change = state_change_observer
    assert state_change_observer in state.callbacks(key='foo')

    state = State({'foo': 'bar'})
    state.callbacks(key='foo').append(state_change_observer)
    state['foo'] = 'spam'
    state_change_observer.assert_called_once_with(
        new_value='spam', old_value='bar', action='update'
    )
    state_change_observer.reset_mock()

    state = State({'foo': 'bar'})
    state.callbacks(key='foo').append(state_change_observer)
    del state['foo']
    state_change_observer.assert_called_once_with(old_value='bar', action='remove')
    state_change_observer.reset_mock()

    state = State({'foo': 'spam'})
    state.callbacks(key='foo').append(state_change_observer)
    state.pop('foo')
    state_change_observer.assert_called_once_with(old_value='spam', action='remove')
    state_change_observer.reset_mock()

    state = State({'foo': 'spam'})
    state.callbacks(key='foo').append(state_change_observer)
    state.popitem()
    state_change_observer.assert_called_once_with(old_value='spam', action='remove')
    state_change_observer.reset_mock()

    state = State({'foo': 'spam'})
    state.callbacks(key='foo').append(state_change_observer)
    state.clear()
    state_change_observer.assert_called_once_with(old_value='spam', action='remove')
    state_change_observer.reset_mock()

    # Update Dict: value gets updated
    state = State({'foo': 'spam'})
    state.callbacks(key='foo').append(state_change_observer)
    state.update({'foo': 'eggs'})
    state_change_observer.assert_called_once_with(
        new_value='eggs', old_value='spam', action='update'
    )
    state_change_observer.reset_mock()

    # Update Dict: value does not get updated
    state = State({'bar': 'spam'})
    state.callbacks(key='foo').append(state_change_observer)
    state.update({'foo': 'eggs'})
    state_change_observer.assert_called_once_with(new_value='eggs', action='add')
    state_change_observer.reset_mock()

    # Update Iterable: value gets updated
    state = State({'foo': 'spam'})
    state.callbacks(key='foo').append(state_change_observer)
    state.update([('foo', 'eggs')])
    state_change_observer.assert_called_once_with(
        new_value='eggs', old_value='spam', action='update'
    )
    state_change_observer.reset_mock()

    # Update Iterable: value does not get updated
    state = State({'bar': 'spam'})
    state.callbacks(key='foo').append(state_change_observer)
    state.update([('foo', 'eggs')])
    state_change_observer.assert_called_once_with(new_value='eggs', action='add')
    state_change_observer.reset_mock()

    # Update Kwargs: value gets updated
    state = State({'foo': 'spam'})
    state.callbacks(key='foo').append(state_change_observer)
    state.update(foo='eggs')
    state_change_observer.assert_called_once_with(
        new_value='eggs', old_value='spam', action='update'
    )
    state_change_observer.reset_mock()

    # Update Kwargs: value does not get updated
    state = State({'bar': 'spam'})
    state.callbacks(key='foo').append(state_change_observer)
    state.update(foo='eggs')
    state_change_observer.assert_called_once_with(new_value='eggs', action='add')
    state_change_observer.reset_mock()

    state = State({'foo': 'spam'})
    state.callbacks(key='foo').append(state_change_observer)
    state.setdefault('foo', 'eggs')
    state_change_observer.assert_not_called()
    state_change_observer.reset_mock()

    state = State({})
    state.callbacks(key='foo').append(state_change_observer)
    state.setdefault('foo', 'eggs')
    state_change_observer.assert_called_once_with(new_value='eggs', action='add')
    state_change_observer.reset_mock()


def test_update_is_not_sent_if_old_value_matches_new_one() -> None:
    """
    - __setitem__
    - __delitem__
    - pop
    - popitem
    - clear
    - update - dict
    - update - iterable
    - update - kwargs
    - setdefault
    """
    state_change_observer = MagicMock()

    state = State({'foo': 'bar'})
    state.callbacks(key='foo').append(state_change_observer)
    state['foo'] = 'bar'
    state_change_observer.assert_not_called()
    state_change_observer.reset_mock()

    # # Dict
    state = State({'foo': 'bar'})
    state.callbacks(key='foo').append(state_change_observer)
    state.update({'foo': 'bar'})
    state_change_observer.assert_not_called()
    state_change_observer.reset_mock()

    # # Iterable
    state = State({'foo': 'bar'})
    state.callbacks(key='foo').append(state_change_observer)
    state.update([('foo', 'bar')])
    state_change_observer.assert_not_called()
    state_change_observer.reset_mock()

    # # Kwargs
    state = State({'foo': 'bar'})
    state.callbacks(key='foo').append(state_change_observer)
    state.update(foo='bar')
    state_change_observer.assert_not_called()
    state_change_observer.reset_mock()


def test_nested_dict_observation() -> None:
    state_change_observer = MagicMock()
    state = State(
        {
            'foo': {'bar': 'spam'},
        }
    )
    state['foo'].callbacks(key='bar').append(state_change_observer)
    state['foo']['bar'] = 'eggs'

    state_change_observer.assert_called_once_with(
        new_value='eggs', old_value='spam', action='update'
    )
    state_change_observer.reset_mock()

    # 3 level nesting
    state_change_observer = MagicMock()
    state = State(
        {
            'one': {'two': {'three': 'four'}},
        }
    )
    state['one']['two'].callbacks(key='three').append(state_change_observer)
    state['one']['two']['three'] = 'five'

    state_change_observer.assert_called_once_with(
        new_value='five', old_value='four', action='update'
    )
    state_change_observer.reset_mock()


def test_add_callback_for_non_existing_key() -> None:
    state_change_observer = MagicMock()
    state = State(
        {
            'foo': {'bar': 'spam'},
        }
    )
    state['foo'].callbacks(key='bar').append(state_change_observer)
    state['foo']['bar'] = 'eggs'

    state_change_observer.assert_called_once_with(
        new_value='eggs', old_value='spam', action='update'
    )
    state_change_observer.reset_mock()

    # 3 level nesting
    state_change_observer = MagicMock()
    state = State(
        {
            'one': {'two': {'three': 'four'}},
        }
    )
    state['one'].callbacks(key='three').append(state_change_observer)
    state['one']['three'] = 'five'

    state_change_observer.assert_called_once_with(new_value='five', action='add')
    state_change_observer.reset_mock()



import json
from logging import getLogger

import httpx

from autogit.constants import PullRequestStates
from autogit.data_types import HttpRequestParams, RepoState
from autogit.utils.helpers import get_access_token
from autogit.utils.throttled_tasks_executor import ThrottledTasksExecutor

logger = getLogger()


def get_http_request_params_for_pull_request_creation(
    repo: RepoState,
) -> HttpRequestParams:
    """Gitlab create MR docs: https://docs.gitlab.com/ee/api/merge_requests.html#create-mr
    Github create MR docs: https://docs.github.com/en/rest/pulls/pulls?apiVersion=2022-11-28#create-a-pull-request.
    """
    if repo.domain == 'github.com':
        # TODO: add group support
        url = f'https://api.github.com/repos/{repo.owner}/{repo.name}/pulls'
        headers = {
            'Accept': 'application/vnd.github+json',
            'Authorization': f'Bearer {get_access_token(repo.url)}',
            'X-GitHub-Api-Version': '2022-11-28',
        }
        data = {
            'title': repo.args.commit_message,
            'body': repo.args.commit_message,
            'head': repo.branch,
            'base': repo.target_branch,
        }

    else:  # Use gitlab.com API by default
        project_id = f'{repo.group + "%2F" if repo.group else ""}{repo.owner}%2F{repo.name}'
        url = f'https://{repo.domain}/api/v4/projects/{project_id}/merge_requests'
        headers = {'PRIVATE-TOKEN': get_access_token(repo.url)}
        data = {
            'source_branch': repo.branch,
            'target_branch': repo.target_branch,
            'title': repo.args.commit_message,
        }
    return HttpRequestParams(
        url=url,
        headers=headers,
        data=data,
    )


def print_pull_requests(repos):
    print('\n\033[1;32m' + 'Created Pull Requests'.center(79, ' ') + '\033[0m')
    show_not_created_pull_requests = False
    for repo in repos.values():
        if repo.pull_request_state == PullRequestStates.CREATED.value:
            print(f'\033[1;32m\033[0m {repo.pull_request_url.ljust(77, " ")} \033[1;32m\033[0m')
        else:
            show_not_created_pull_requests = True
    if show_not_created_pull_requests:
        print('\033[1;32m' + 'Not created Pull Requests'.center(79, ' ') + '\033[0m')
        for repo in repos.values():
            if repo.pull_request_state == PullRequestStates.GOT_BAD_RESPONSE.value:
                print(f'\033[1;32m\033[0m {repo.url.ljust(77, " ")} \033[1;32m\033[0m')
                print(
                    f'\033[1;32m\033[0m   status_code={repo.pull_request_status_code} reason={repo.pull_request_reason} \033[1;32m\033[0m'
                )


async def create_pull_request(repo: RepoState) -> None:
    # https://stackoverflow.com/questions/56027634/creating-a-pull-request-using-the-api-of-github
    request_params = get_http_request_params_for_pull_request_creation(repo)

    async with httpx.AsyncClient() as client:
        response = await client.post(
            url=request_params.url,
            headers=request_params.headers,
            data=request_params.data,
        )
        if response.status_code < 400:
            repo.pull_request_state = PullRequestStates.CREATED.value
            repo.pull_request_url = response.json().get('web_url', response.json().get('url'))
        else:
            repo.pull_request_state = PullRequestStates.GOT_BAD_RESPONSE.value
            repo.pull_request_status_code = response.status_code
            repo.pull_request_reason = json.dumps(response.json())


def create_pull_request_for_each_repo(
    repos: dict[str, RepoState], executor: ThrottledTasksExecutor
) -> None:
    for repo in repos.values():
        executor.run(create_pull_request(repo))
    executor.wait_for_tasks_to_finish()
    print_pull_requests(repos)



import git

from autogit.constants import ModificationState
from autogit.data_types import RepoState
from autogit.utils.throttled_tasks_executor import ThrottledTasksExecutor


async def commit_and_push_changes(repo: RepoState) -> None:
    g = git.Repo(repo.directory)

    if g.index.diff(None) or g.untracked_files:
        g.git.add(A=True)
        g.git.commit(m=repo.args.commit_message)

        g.git.push('--set-upstream', 'origin', repo.branch)
        repo.modification_state = ModificationState.PUSHED_TO_REMOTE.value
    else:
        repo.modification_state = ModificationState.NO_FILES_CHANGED.value


def print_modified_repositories(repos: dict[str, RepoState]):
    branch = next(iter(repos.values())).branch
    print('\n\033[1;32m' + f'Created branch: {branch}'.center(79, ' ') + '\033[0m')
    should_print_not_modified_repos = False
    print_repo_exceptions = False
    for repo in repos.values():
        if repo.modification_state == ModificationState.PUSHED_TO_REMOTE.value:
            print(f'\033[1;32m\033[0m {repo.url.ljust(77, " ")} \033[1;32m\033[0m')
        else:
            if repo.modification_state == ModificationState.GOT_EXCEPTION.value:
                print_repo_exceptions = True
            should_print_not_modified_repos = True

    if should_print_not_modified_repos:
        print(
            '\033[1;32m|\033[0m'
            + 'Did NOT modify these repositories:'.center(77, ' ')
            + '\033[1;32m|\033[0m'
        )
        for repo in repos.values():
            if repo.cloning_state != ModificationState.PUSHED_TO_REMOTE.value:
                print(
                    f'\033[1;32m|\033[0m {(repo.url + " " + repo.modification_state).ljust(75, " ")} \033[1;32m|\033[0m'
                )

    if print_repo_exceptions:
        print('\033[1;32m\033[0m' + 'Exceptions:'.center(79, ' ') + '\033[1;32m\033[0m')
        for repo in repos.values():
            if repo.cloning_state == ModificationState.GOT_EXCEPTION.value:
                print(
                    '\033[1;32m\033[0m'
                    f' {(repo.url + " " + repo.modification_state).ljust(77, " ")}:'
                    '\033[1;32m\033[0m'
                )
                print(repo.stderr)


def commit_and_push_changes_for_each_repo(
    repos: dict[str, RepoState], executor: ThrottledTasksExecutor
) -> None:
    for repo in repos.values():
        executor.run(commit_and_push_changes(repo))
    executor.wait_for_tasks_to_finish()
    print_modified_repositories(repos)



"""Store and modify global app state in JSON (Python dictionary) to re-render front-end components."""

from collections.abc import Callable, Hashable, Iterable, Iterator, MutableMapping
from typing import Any, SupportsIndex, overload

__version__ = '0.0.6'


class DictState(dict):
    def __init__(self, json_state: dict | MutableMapping, *args, **kwargs) -> None:
        self.on_change_callbacks = {}

        # Convert all nested JSON to State
        if isinstance(json_state, dict | MutableMapping):
            for key in json_state:
                value = json_state[key]
                if isinstance(value, dict | MutableMapping | list):
                    json_state[key] = State(value)

        super().__init__(json_state, *args, **kwargs)

    def callbacks(self, key: str) -> list[Callable]:
        self.on_change_callbacks.setdefault(key, [])
        return self.on_change_callbacks[key]

    def pop(
        self, key: Any, default: Any | None = None, /, *args, **kwargs
    ) -> Any | None:
        if key not in self:
            return super().pop(key, default, *args, **kwargs)

        old_value = self[key]
        result = super().pop(key, default, *args, **kwargs)
        for callback in self.on_change_callbacks.get(key, []):
            callback(old_value=old_value, action='remove')
        return result

    def popitem(self) -> tuple[Any, Any]:
        result = super().popitem()
        if result:
            key = result[0]
            old_value = result[1]
            for callback in self.on_change_callbacks.get(key, []):
                callback(old_value=old_value, action='remove')
        return result

    def clear(self) -> None:
        for key in self:
            for callback in self.on_change_callbacks.get(key, []):
                callback(old_value=self[key], action='remove')
        return super().clear()

    @overload
    def update(self, arg: MutableMapping[Hashable, Any], /, **kwargs: Any) -> None: ...

    @overload
    def update(self, arg: Iterable[tuple[Hashable, Any]], /, **kwargs: Any) -> None: ...

    @overload
    def update(self, /, **kwargs: Any) -> None: ...

    def update(self, arg: Any = None, /, **kwargs: Any) -> None:  # type: ignore
        if hasattr(arg, 'keys'):
            for key in arg:
                # TODO: do not send update callbacks if old and new values are the same.
                new_value = arg[key]
                callback_kwargs = None
                if key in self:
                    old_value = self[key]
                    if old_value != new_value:
                        callback_kwargs = {
                            'new_value': new_value,
                            'old_value': old_value,
                            'action': 'update',
                        }
                else:
                    callback_kwargs = {
                        'new_value': new_value,
                        'action': 'add',
                    }

                if callback_kwargs:
                    for callback in self.on_change_callbacks.get(key, []):
                        callback(**callback_kwargs)

            return super().update(arg, **kwargs)
        if arg:
            for key, value in arg:
                new_value = value
                callback_kwargs = None
                if key in self:
                    old_value = self[key]
                    if old_value != new_value:
                        callback_kwargs = {
                            'new_value': new_value,
                            'old_value': old_value,
                            'action': 'update',
                        }
                else:
                    callback_kwargs = {
                        'new_value': new_value,
                        'action': 'add',
                    }

                if callback_kwargs:
                    for callback in self.on_change_callbacks.get(key, []):
                        callback(**callback_kwargs)
            return super().update(arg, **kwargs)
        for key, value in kwargs.items():
            new_value = value
            callback_kwargs = None
            if key in self:
                old_value = self[key]
                if old_value != new_value:
                    callback_kwargs = {
                        'new_value': new_value,
                        'old_value': old_value,
                        'action': 'update',
                    }
            else:
                callback_kwargs = {
                    'new_value': new_value,
                    'action': 'add',
                }

            if callback_kwargs:
                for callback in self.on_change_callbacks.get(key, []):
                    callback(**callback_kwargs)

        return super().update(kwargs, **kwargs)

    def setdefault(self, key, default: Any | None = None, /) -> Any | None:
        callback_kwargs = None
        if key not in self:
            callback_kwargs = {
                'action': 'add',
            }

        result = super().setdefault(key, default)

        if callback_kwargs:
            for callback in self.on_change_callbacks.get(key, []):
                callback(new_value=result, **callback_kwargs)

        return result

    def __setitem__(self, key: Any, value: Any, /, *args, **kwargs) -> None:
        callback_kwargs = None
        new_value = value
        if key in self:
            old_value = self[key]
            if old_value != new_value:
                callback_kwargs = {
                    'new_value': new_value,
                    'old_value': self[key],
                    'action': 'update',
                }
        else:
            callback_kwargs = {
                'new_value': new_value,
                'action': 'add',
            }

        result = super().__setitem__(key, value, *args, **kwargs)

        if callback_kwargs:
            for callback in self.on_change_callbacks.get(key, []):
                callback(**callback_kwargs)

        return result

    def __delitem__(self, key: Any, /) -> None:
        if key not in self:
            return super().__delitem__(key)

        old_value = self[key]
        result = super().__delitem__(key)
        for callback in self.on_change_callbacks.get(key, []):
            callback(old_value=old_value, action='remove')
        return result


class ListState(list):
    """Events emitted:
    add - element is added to a list.
    remove - element is removed from the list.
    update - element is replaced.
    order - place of an element in list is changed.  When element is removed - the event wont be emitted.
    """

    def __init__(self, json_state: list, *args, **kwargs) -> None:
        self.on_change_callbacks = []

        # Convert all nested JSON to State
        if isinstance(json_state, list):
            for index, value in enumerate(json_state):
                if isinstance(value, dict | MutableMapping | list):
                    json_state[index] = State(value)

        super().__init__(json_state, *args, **kwargs)

    def callbacks(self) -> list[Callable]:
        return self.on_change_callbacks

    def append(self, object: Any, /, **kwargs) -> None:
        super().append(object)
        for callback in self.on_change_callbacks:
            callback(new_value=object, action='add')

    def __setitem__(
        self, key: SupportsIndex, value: Any | Iterator[Any], /, **kwargs
    ) -> None:  # type: ignore
        old_value = self[key]
        super().__setitem__(key, value, **kwargs)
        # TODO: check if this key is in a list
        for callback in self.on_change_callbacks:
            callback(new_value=value, old_value=old_value, action='update')

    def __delitem__(self, key: SupportsIndex | slice, /, **kwargs) -> None:
        old_value = self[key]
        # TODO: check if the key is in the list
        # TODO: support slice
        super().__delitem__(key, **kwargs)
        for callback in self.on_change_callbacks:
            callback(old_value=old_value, action='remove')

    def pop(self, index: SupportsIndex = -1, /, **kwargs) -> None:
        # TODO: check if this index is in a list
        old_value = self[index]
        super().pop(index, **kwargs)
        for callback in self.on_change_callbacks:
            callback(old_value=old_value, action='remove')

    def extend(self, iterable, /, **kwargs) -> None:
        elements = list(iterable)
        super().extend(elements, **kwargs)
        for element in elements:
            for callback in self.on_change_callbacks:
                callback(new_value=element, action='add')

    def insert(self, index: SupportsIndex, object: Any, /, **kwargs) -> None:
        super().insert(index, object, **kwargs)
        for callback in self.on_change_callbacks:
            callback(
                new_value=object, action='add', index=index
            )  # TODO: add `index` argument for list updates. Index changes should also be observable.

    def remove(self, value: Any, /, **kwargs) -> None:
        is_value_in_list = value in self
        super().remove(value, **kwargs)
        if is_value_in_list:
            for callback in self.on_change_callbacks:
                callback(old_value=value, action='remove')

    def clear(self) -> None:
        for element in self:
            for callback in self.on_change_callbacks:
                callback(old_value=element, action='remove')

        super().clear()

    # TODO: support ordering: update the index, action="order", value=<current_value>, index- element index in the list.

    def reverse(self) -> None:
        super().reverse()

    def __reversed__(self) -> Iterator[Any]:
        return super().__reversed__()

    def sort(self, *, key: Any | None = None, reverse: bool = False) -> None:
        super().sort(key=key, reverse=reverse)

    def copy(self) -> list[Any]:
        return super().copy()


class SetState(set):
    # TODO: add implementation
    pass


class State(type):
    def __new__(
        cls, json_state: dict | list, *args, **kwargs
    ) -> DictState | ListState | SetState:
        if isinstance(json_state, dict):
            return DictState(json_state)
        if isinstance(json_state, list):
            return ListState(json_state)
        if isinstance(json_state, set):
            return SetState(json_state)
        return None



import os
import subprocess

from autogit.data_types import ModificationState, RepoState
from autogit.utils.throttled_tasks_executor import ThrottledTasksExecutor


async def run_command(repo: RepoState) -> None:
    commands = repo.args.commands

    # Expand the name of the first argument if its a file and its in current directory
    if commands and os.path.exists(commands[0]) and os.path.isfile(commands[0]):
        commands[0] = os.path.abspath(commands[0])

    # Execute commands
    proc = subprocess.Popen(
        repo.args.commands,
        cwd=os.path.abspath(repo.directory),
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
    )
    # TODO: Show output in real time:  https://stackoverflow.com/a/20576150
    # TODO: log the output in the temporal files.
    repo.stdout, repo.stderr = proc.communicate()
    if proc.returncode:
        repo.modification_state = ModificationState.GOT_EXCEPTION.value
    else:
        repo.modification_state = ModificationState.MODIFIED.value


def run_command_for_each_repo(
    repos: dict[str, RepoState], executor: ThrottledTasksExecutor
) -> None:
    for repo in repos.values():
        executor.run_not_throttled(run_command(repo))
    executor.wait_for_tasks_to_finish()



#!/usr/bin/env python

import glob
import os
import re
from typing import List, Optional


def get_all_file_paths(directory: Optional[str] = None) -> List[str]:
    """Provides a list of file-paths for every file in a directory (including subdirectories)."""
    if directory is None:
        directory = os.getcwd()

    recursive_directory_path = os.path.abspath(f"{directory.rstrip('/')}/**")
    file_and_dir_paths = glob.glob(recursive_directory_path, recursive=True)
    return [path for path in file_and_dir_paths if not path.endswith("/")]


def replace(filename: str, pattern: str, replace_with: str) -> None:
    """Replaces all patterns found in a file with a `replace_with` fragment."""
    with open(filename, "r") as f:
        content = f.read()

    for mo in list(re.finditer(pattern, content))[::-1]:
        content = content[:mo.start()] + replace_with + content[mo.end():]

    with open(filename, "w") as f:
        f.write(content)


if __name__ == "__main__":
    for filename in get_all_file_paths():
        replace(filename, pattern="mypy==[\\d.]*", replace_with="mypy==0.991")



from dataclasses import dataclass

from autogit.constants import CloningStates, ModificationState, PullRequestStates


@dataclass
class CliArguments:
    action_id: str  # Generated hash for action identification
    repos: list[str]  # A list of Urls or files containing Urls
    clone_to: str  # Directory which will be used to clone repos to
    commands: list[str]  # Commands which have to be exeucted in cloned repo
    commit_message: str  # Message which will be used for commit, branch, PR (if not provided)
    verbose: bool  # Provides additional debug information
    source_branch: (
        str | None
    )  # Base branch on which the new branch will be created (if it does not exist yet)
    branch: (
        str | None
    )  # Branch to use in PR as a source branch (it will be created if does not exist)
    target_branch: str | None  # Branch to use in PR as a target branch


@dataclass
class RepoState:
    args: CliArguments  # Parsed command line arguments

    source_branch: str = ''  # Branch name from which a new branch for changes will be created
    branch: str = ''  # Branch name in which changes will be made and commited
    target_branch: str = ''  # Base branch into which PR changes will be pulled

    cloning_state: str = CloningStates.NOT_STARTED.value
    modification_state: str = ModificationState.NOT_STARTED.value
    pull_request_state: str = PullRequestStates.NOT_CREATED.value
    pull_request_status_code: int | None = None
    pull_request_reason: str | None = None

    name: str = ''  # Short human readable repo identifier
    owner: str = ''  # Owner of this repo
    group: str | None = ''  # Group of this repo
    url: str = ''  # Url used to clone the repository
    domain: str = ''  # Domain where the remote repository is hosted at (parsed from url)
    pull_request_url: str = ''  # Link to created pull request
    directory: str = ''  # Repository path in the file system

    stdout: bytes = b''  # Standard output from command execution
    stderr: bytes = b''  # Standard error output from command execution


@dataclass
class HttpRequestParams:
    url: str
    headers: dict[str, str]
    data: dict[str, str]
    json: dict[str, str] | None = None



import contextlib

import git
from git.cmd import Git

from autogit.data_types import RepoState
from autogit.utils.helpers import to_kebab_case
from autogit.utils.throttled_tasks_executor import ThrottledTasksExecutor


async def create_branch(repo: RepoState) -> None:
    if repo.args.branch:
        new_branch_name = repo.args.branch
    else:
        new_branch_name = to_kebab_case(repo.args.commit_message)
        if repo.args.action_id and repo.args.action_id not in new_branch_name:
            new_branch_name += f'-{repo.args.action_id}'

    repo.branch = new_branch_name

    g = Git(repo.directory)

    # TODO: add a conditional check if the branch exists or not
    # TODO: what should be done if the branch exists and contains changes? Error should be shown and action canceled with nice error message.
    try:
        g.execute(['git', 'checkout', '-b', repo.branch])
    except git.exc.GitCommandError:
        g.execute(['git', 'checkout', repo.branch])

    with contextlib.suppress(git.exc.GitCommandError):
        g.execute(['git', 'pull', 'origin', repo.branch])


def create_branch_for_each_repo(
    repos: dict[str, RepoState], executor: ThrottledTasksExecutor
) -> None:
    for repo in repos.values():
        executor.run_not_throttled(create_branch(repo))
    executor.wait_for_tasks_to_finish()



from unittest.mock import MagicMock, call

from jsonstate import State


def test_callback_for_list_methods_at_root_level() -> None:
    """
    Test observer callback invocations for these List methods:
    '__delitem__',
    '__reversed__',
    '__setitem__',
    'append',
    'clear',
    'copy',
    'extend',
    'insert',
    'pop',
    'remove',
    'reverse',
    'sort'
    """
    state_change_observer = MagicMock()

    state = State(['foo', 'bar'])
    assert state
    assert state[0] == 'foo'
    assert state[1] == 'bar'

    state.callbacks().append(state_change_observer)
    assert state_change_observer in state.callbacks()

    state.append('spam')
    assert 'spam' in state
    state_change_observer.assert_called_once_with(new_value='spam', action='add')
    state_change_observer.reset_mock()

    state[-1] = 'eggs'
    assert 'eggs' in state
    state_change_observer.assert_called_once_with(
        new_value='eggs', old_value='spam', action='update'
    )
    state_change_observer.reset_mock()

    assert 'eggs' in state
    del state[-1]
    state_change_observer.assert_called_once_with(old_value='eggs', action='remove')
    state_change_observer.reset_mock()

    assert 'bar' in state
    state.pop()
    state_change_observer.assert_called_once_with(old_value='bar', action='remove')
    state_change_observer.reset_mock()

    state.extend(['bar', 'spam'])
    state_change_observer.assert_has_calls(
        [call(new_value='bar', action='add'), call(new_value='spam', action='add')]
    )
    state_change_observer.reset_mock()

    state.insert(0, 'FOO')
    assert 'FOO' in state
    state_change_observer.assert_called_once_with(
        new_value='FOO', action='add', index=0
    )
    state_change_observer.reset_mock()

    state.remove('FOO')
    assert 'FOO' not in state
    state_change_observer.assert_called_once_with(old_value='FOO', action='remove')
    state_change_observer.reset_mock()

    state.clear()
    assert state == []
    state_change_observer.assert_has_calls(
        [
            call(old_value='foo', action='remove'),
            call(old_value='bar', action='remove'),
            call(old_value='spam', action='remove'),
        ]
    )
    state_change_observer.reset_mock()


def test_callback_for_list_methods_at_nested_level() -> None:
    state_change_observer = MagicMock()

    state = State(['foo', ['bar']])

    state[1].callbacks().append(state_change_observer)
    assert state_change_observer in state[1].callbacks()

    state[1].append('spam')
    state_change_observer.assert_called_once_with(new_value='spam', action='add')
    state_change_observer.reset_mock()



import os.path
from urllib.parse import urlparse

import git
from git.cmd import Git
from git.exc import GitCommandError

from autogit.constants import CloningStates
from autogit.data_types import RepoState
from autogit.utils.helpers import get_access_token, get_default_branch
from autogit.utils.throttled_tasks_executor import ThrottledTasksExecutor


def get_repo_access_url(url: str) -> str | None:
    """Converts repository url to url which is suitable for cloning."""
    if url.startswith('http'):
        if access_token := get_access_token(url):
            parsed_url = urlparse(url)
            domain_with_access_token = f'api:{access_token}@{parsed_url.netloc.split("@")[-1]}'
            parsed_url = parsed_url._replace(netloc=domain_with_access_token, scheme='https')
            return parsed_url.geturl()
    elif url.startswith('git@'):
        return url
    return None


async def clone_repository(repo: RepoState) -> None:
    """Clones repository with default (or source) branch."""
    clone_to = repo.args.clone_to
    repo.directory = os.path.join(clone_to, repo.name)
    repo.target_branch = repo.target_branch or get_default_branch(repo)
    repo.source_branch = repo.source_branch or repo.target_branch

    # TODO: add a way to clone using access token: https://stackoverflow.com/questions/25409700/using-gitlab-token-to-clone-without-authentication/29570677#29570677
    # git clone https://:YOURKEY@your.gilab.company.org/group/project.git

    # TODO: add ssh support: urls like git@gitlab.com:niekas/gitlab-api-tests.git

    try:
        if os.path.exists(
            repo.directory
        ):  # If repository exists: clean it, pull changes, checkout default branch
            g = Git(repo.directory)
            g.clean('-dfx')
            g.execute(['git', 'fetch', '--all'])

            g.checkout(repo.source_branch)

            repo.cloning_state = CloningStates.CLONED.value
        elif repo_access_url := get_repo_access_url(repo.url):
            git.Repo.clone_from(repo_access_url, repo.directory)

            g = Git(repo.directory)
            g.execute(['git', 'fetch', '--all'])

            source_branch = repo.source_branch or repo.target_branch or get_default_branch(repo)
            # TODO: check if this branch exist - if not - show nice error message and abort whole action
            g.checkout(source_branch)

            repo.cloning_state = CloningStates.CLONED.value
        else:
            repo.cloning_state = CloningStates.ACCESS_TOKEN_NOT_PROVIDED.value

    except GitCommandError:
        repo.cloning_state = CloningStates.NOT_FOUND.value


def print_cloned_repositories(repos):
    # TODO: flush print message after each repository action is done (not to freeze the screen if cloning multiple repositories takes too long)
    clone_to = next(iter(repos.values())).args.clone_to
    print('\n\033[1;32m' + f'Cloned repositories (to {clone_to})'.center(79, ' ') + '\033[0m')
    should_print_not_cloned_repos = False
    for repo in repos.values():
        if repo.cloning_state == CloningStates.CLONED.value:
            print(f'\033[1;32m\033[0m {repo.url.ljust(77, " ")} \033[1;32m\033[0m')
        else:
            should_print_not_cloned_repos = True
    if should_print_not_cloned_repos:
        print('\033[1;32m' + 'Did NOT clone these repositories:'.center(79, '-') + '\033[0m')
        for repo in repos.values():
            if repo.cloning_state != repo.cloning_state:
                print(
                    f'\033[1;32m\033[0m {(repo.url + " " + CloningStates.CLONED.value).ljust(77, " ")} \033[1;32m\033[0m'
                )


def clone_repositories(repos: dict[str, RepoState], executor: ThrottledTasksExecutor) -> None:
    for repo in repos.values():
        executor.run(clone_repository(repo))
    executor.wait_for_tasks_to_finish()
    print_cloned_repositories(repos)



import asyncio
import time
import traceback
from asyncio import Future
from collections.abc import Callable, Coroutine
from threading import Thread


class ThrottledTasksExecutor:
    """Executor which allows to run coroutines without hitting throttling limits.

    Usage example:
        async def generate_greeting(name: str) -> str:
            from datetime import datetime
            print(f"Started greeting generation for {name} at {datetime.now()}")
            await asyncio.sleep(1)
            print(f"Done greeting generation for {name} at {datetime.now()}")
            return f"Hello, {name}!"

        def process_result(greeting: str) -> None:
            print(f"This result was generated: {greeting}")

        with ThrottledTasksExecutor(delay_between_tasks=2) as executor:
            executor.run(generate_greeting("World"), callback=process_result)
            executor.run(generate_greeting("Universe"), callback=process_result)
    """

    def __init__(
        self, delay_between_tasks: float = 0.2, in_separate_process: bool = False
    ) -> None:
        self.loop = asyncio.new_event_loop()
        self.running_tasks: set[Future] = (
            set()
        )  # TODO: Find the example with task removal from the set.
        # TODO: Would be nice to have awaitable future, instead of infinite loop.
        self.delay_between_tasks = delay_between_tasks
        self.in_separate_process = in_separate_process

        self.can_task_be_executed = asyncio.Condition()
        self.is_running = False

    def __enter__(self) -> 'ThrottledTasksExecutor':
        self.start()
        return self

    def __exit__(self, *exc_info) -> None:
        self.wait_for_tasks_to_finish()
        self.stop()

    def start(self, in_separate_process: bool | None = None) -> None:
        """Starts a thread (or a process), which executes coroutines provided to the ThrottledTasksExecutor."""
        if in_separate_process is None:
            in_separate_process = self.in_separate_process

        if in_separate_process:
            # TODO: investigate a way to start coroutines in a separate process:
            #   - https://docs.python.org/3/library/asyncio-eventloop.html#asyncio.loop.run_in_executor
            msg = 'Running executor in a separate process is not supported yet'
            raise NotImplementedError(msg)
        thread = Thread(target=self._run_event_loop, daemon=True)
        thread.start()

        self.is_running = True

        # Periodically emit permission to execute one task
        self._allow_task_execution_task = asyncio.run_coroutine_threadsafe(
            self._allow_task_execution(every=self.delay_between_tasks), self.loop
        )

    def stop(self) -> None:
        """Terminates a thread (or a process), which executes coroutines provided to the ThrottledTasksExecutor."""
        self.is_running = False
        self._allow_task_execution_task.cancel()

        time.sleep(self.delay_between_tasks)  # Wait for _allow_task_execution_task to be destroyed
        self.loop.stop()

    def run(
        self,
        coroutine: Coroutine | Callable,
        *args,
        callback: Callable | None = None,
        **kwargs,
    ) -> None:
        """Executes coroutine in an executor thread, which makes sure not to hit throttling limits."""
        if callback is None:

            def callback(*args, **kwargs):
                return None

        if not isinstance(coroutine, Coroutine):
            coroutine = coroutine(*args, **kwargs)
        if not isinstance(coroutine, Coroutine):
            msg = 'Can only execute coroutines, not coroutine provided'
            raise ValueError(msg)

        task: Future = asyncio.run_coroutine_threadsafe(  # type: ignore
            self._throttled_task(coroutine), self.loop
        )
        self.running_tasks.add(task)
        task.add_done_callback(self._mark_task_done(callback))

    def run_not_throttled(
        self,
        coroutine: Coroutine | Callable,
        *args,
        callback: Callable | None = None,
        **kwargs,
    ) -> None:
        """Executes coroutine in an executor event loop ignoring throttled tasks queue."""
        if callback is None:

            def callback(*args, **kwargs):
                return None

        if not isinstance(coroutine, Coroutine):
            coroutine = coroutine(*args, **kwargs)
        if not isinstance(coroutine, Coroutine):
            msg = 'Can only execute coroutines, not coroutine provided'
            raise ValueError(msg)

        task: Future = asyncio.run_coroutine_threadsafe(coroutine, self.loop)  # type: ignore
        self.running_tasks.add(task)
        task.add_done_callback(self._mark_task_done(callback))

    def wait_for_tasks_to_finish(self) -> None:
        asyncio.run(self.async_wait_for_tasks_to_finish())

    async def async_wait_for_tasks_to_finish(self) -> None:
        while self.running_tasks:
            await asyncio.sleep(0.1)

    async def _allow_task_execution(self, every: float, count: int = 1) -> None:
        """Periodically emits event, which allows for `count` tasks to be executed.

        Params:
            every: float  - seconds how often a permission is given to `count` tasks to be executed.
            count: int    - how many tasks are allowed to be run at each event.
        """
        try:
            while self.is_running:
                # Does the first task has to wait `every` seconds to be started?
                async with self.can_task_be_executed:
                    if count and count > 0:
                        self.can_task_be_executed.notify(n=count)
                    else:
                        self.can_task_be_executed.notify_all()
                await asyncio.sleep(every)
        except asyncio.CancelledError:
            pass

    def _throttled_task(self, coroutine: Coroutine) -> Coroutine:
        """Decorator for coroutine to wait for permission before executing the coroutine."""

        async def throttled_task_wrapper(*args, **kwargs):
            async with self.can_task_be_executed:
                await self.can_task_be_executed.wait()
            return await coroutine

        return throttled_task_wrapper()

    def _mark_task_done(self, callback) -> Callable:
        """Decorator for callback to set the task as done after the callback is processed."""

        def task_done_wrapper(task: Future) -> None:
            try:
                task_result = task.result()  # type: ignore
            except Exception:
                traceback.print_exc()
            else:
                try:
                    callback(task_result)
                except Exception:
                    traceback.print_exc()
            self.running_tasks.discard(task)

        return task_done_wrapper

    def _run_event_loop(self) -> None:
        asyncio.set_event_loop(self.loop)
        self.loop.run_forever()



from unittest.mock import MagicMock


def test_basic_usage_flow() -> None:
    """This example is provided in the README.md"""

    from jsonstate import State

    state = State(
        {
            'title': 'State Example',
            'profile': {
                'name': 'Foo',
            },
            'products': [
                {'name': 'Foo', 'description': 'Foo spam'},
                {'name': 'Bar', 'description': 'Bar spam'},
            ],
        }
    )
    # print_event = lambda **kwargs: print("Event", kwargs)
    print_event = MagicMock()

    state.callbacks(key='title').append(print_event)
    state['profile'].callbacks(key='name').append(print_event)
    state['products'].callbacks().append(print_event)

    # This statement updates the state and also invokes on_change callback:
    state['title'] = 'Eggs'
    # Event {'new_value': 'Eggs', 'old_value': 'State Example', 'action': 'update'}
    print_event.assert_called_once_with(
        new_value='Eggs', old_value='State Example', action='update'
    )
    print_event.reset_mock()

    # This statement updates the state and also invokes on_change callback:
    state['profile']['name'] = 'Spam'
    # Event {'new_value': 'Eggs', 'old_value': 'State Example', 'action': 'update'}
    print_event.assert_called_once_with(
        new_value='Spam', old_value='Foo', action='update'
    )
    print_event.reset_mock()

    # This statement also updates the state and invokes on_change callback:
    state['products'].append({'name': 'Eggs', 'description': 'Eggs spam'})
    # Event {'new_value': {"name": "Eggs", "description": "Eggs spam"}, 'action': 'append'}

    ## TODO:
    print_event.assert_called_once_with(
        new_value={'name': 'Eggs', 'description': 'Eggs spam'}, action='add'
    )
    # Event {"new_value": {"name": "Eggs", "description": "Eggs spam"}, "action": "add"}
    print_event.reset_mock()
